#include <linux/if_bridge.h>

#include <netlink-private/netlink.h>
#include <netlink/hashtable.h>
#include <netlink/route/bridge_vlan.h>

/**
 * @ingroup bridge
 * @defgroup bridge_vlan Bridge VLAN
 * Bridge vlan operations
 *
 * @route_doc{bridge_vlan, Bridge VLAN Documentation}
 *
 * @details
 * @{
 */

/** @cond SKIP */
#define BRIDGE_VLAN_ATTR_IFINDEX (1 << 0)
#define BRIDGE_VLAN_ATTR_VID (1 << 1)
#define BRIDGE_VLAN_ATTR_FLAGS (1 << 2)
#define BRIDGE_VLAN_ATTR_STATE (1 << 3)
#define BRIDGE_VLAN_ATTR_TUNID (1 << 4)
#define BRIDGE_VLAN_ATTR_RX_BYTES (1 << 5)
#define BRIDGE_VLAN_ATTR_RX_PACKETS (1 << 6)
#define BRIDGE_VLAN_ATTR_TX_BYTES (1 << 6)
#define BRIDGE_VLAN_ATTR_TX_PACKETS (1 << 7)

static struct nl_cache_ops rtnl_bridge_vlan_ops;
static struct nl_object_ops bridge_vlan_obj_ops;
/** @endcond */

static struct nla_policy bridge_vlan_policy[BRIDGE_VLANDB_ENTRY_MAX + 1] = {
	[BRIDGE_VLANDB_ENTRY_INFO] = {
		.minlen = sizeof(struct bridge_vlan_info),
		.maxlen = sizeof(struct bridge_vlan_info),
	},
	[BRIDGE_VLANDB_ENTRY_RANGE] = { .type = NLA_U16 },
	[BRIDGE_VLANDB_ENTRY_STATE] = { .type = NLA_U8 },
	[BRIDGE_VLANDB_ENTRY_TUNNEL_INFO] = { .type = NLA_NESTED },
	[BRIDGE_VLANDB_ENTRY_STATS] = { .type = NLA_NESTED },
};

static struct nla_policy
bridge_vlan_tinfo_policy[BRIDGE_VLANDB_TINFO_MAX + 1] = {
	[BRIDGE_VLANDB_TINFO_ID] = { .type = NLA_U32 },
	[BRIDGE_VLANDB_TINFO_CMD] = { .type = NLA_U32 },
};

static int bridge_vlan_set_vlan_attrs(struct rtnl_bridge_vlan *bv,
				      struct nlattr * const *vlan_tb,
				      uint16_t offset)
{
	if (vlan_tb[BRIDGE_VLANDB_ENTRY_STATE]) {
		bv->bv_state = nla_get_u8(
			vlan_tb[BRIDGE_VLANDB_ENTRY_STATE]);
		bv->ce_mask |= BRIDGE_VLAN_ATTR_STATE;
	}

	if (vlan_tb[BRIDGE_VLANDB_ENTRY_TUNNEL_INFO]) {
		struct nlattr *tun_tb[BRIDGE_VLANDB_TINFO_MAX + 1];
		int err;

		err = nla_parse_nested(tun_tb, BRIDGE_VLANDB_TINFO_MAX,
				       vlan_tb[BRIDGE_VLANDB_ENTRY_TUNNEL_INFO],
				       bridge_vlan_tinfo_policy);
		if (err < 0) {
			nl_perror(err, "Error while parsing BRIDGE_VLANDB_ENTRY_TUNNEL_INFO message");
			return err;
		}

		if (tun_tb[BRIDGE_VLANDB_TINFO_ID]) {
			bv->bv_tunid =
				nla_get_u32(tun_tb[BRIDGE_VLANDB_TINFO_ID]) +
				offset;
			bv->ce_mask |= BRIDGE_VLAN_ATTR_TUNID;
		}
	}

	if (vlan_tb[BRIDGE_VLANDB_ENTRY_STATS]) {
		struct nlattr *stats_tb[BRIDGE_VLANDB_STATS_MAX + 1];
		int err;

		err = nla_parse_nested(stats_tb, BRIDGE_VLANDB_STATS_MAX,
				       vlan_tb[BRIDGE_VLANDB_ENTRY_STATS],
				       NULL);
		if (err < 0) {
			nl_perror(err, "Error while parsing BRIDGE_VLANDB_ENTRY_STATS message");
			return err;
		}

#define BRIDGE_VLAN_GET_STAT(attrname, suffix) \
		if (stats_tb[BRIDGE_VLANDB_STATS_##suffix]) { \
			bv->attrname = \
				nla_get_u64(stats_tb[BRIDGE_VLANDB_STATS_##suffix]); \
			bv->ce_mask |= BRIDGE_VLAN_ATTR_##suffix; \
		}

		BRIDGE_VLAN_GET_STAT(bv_rx_bytes, RX_BYTES);
		BRIDGE_VLAN_GET_STAT(bv_rx_packets, RX_PACKETS);
		BRIDGE_VLAN_GET_STAT(bv_tx_bytes, TX_BYTES);
		BRIDGE_VLAN_GET_STAT(bv_tx_packets, TX_PACKETS);

#undef BRIDGE_VLAN_GET_STAT
	}

	return 0;
}

static int bridge_vlan_msg_parser(struct nl_cache_ops *ops,
				  struct sockaddr_nl *who,
				  struct nlmsghdr *nlh,
				  struct nl_parser_param *pp)
{
	struct br_vlan_msg *bvm;
	struct nlattr *attr;
	int rem;

	if (!nlmsg_valid_hdr(nlh, sizeof(struct br_vlan_msg)))
		return -NLE_MSG_TOOSHORT;
	bvm = nlmsg_data(nlh);

	if (bvm->family != AF_BRIDGE)
		return -NLE_AF_NOSUPPORT;

	nlmsg_for_each_attr(attr, nlh, sizeof(struct br_vlan_msg), rem) {
		struct nlattr *vlan_tb[BRIDGE_VLANDB_ENTRY_MAX + 1];
		struct bridge_vlan_info *vinfo;
		uint16_t vid, range_end;
		int err;

		if (nla_type(attr) != BRIDGE_VLANDB_ENTRY)
			continue;

		err = nla_parse_nested(vlan_tb, BRIDGE_VLANDB_ENTRY_MAX, attr,
				       bridge_vlan_policy);
		if (err < 0) {
			nl_perror(err, "Error while parsing BRIDGE_VLANDB_ENTRY message");
			return err;
		}

		if (!vlan_tb[BRIDGE_VLANDB_ENTRY_INFO])
			continue;
		vinfo = nla_data(vlan_tb[BRIDGE_VLANDB_ENTRY_INFO]);

		if (vlan_tb[BRIDGE_VLANDB_ENTRY_RANGE]) {
			range_end =
				nla_get_u16(vlan_tb[BRIDGE_VLANDB_ENTRY_RANGE]);
		} else {
			range_end = vinfo->vid;
		}

		for (vid = vinfo->vid; vid <= range_end; vid++) {
			struct rtnl_bridge_vlan *bv;

			bv = rtnl_bridge_vlan_alloc();
			if (!bv)
				return -NLE_NOMEM;

			bv->ce_msgtype = nlh->nlmsg_type;

			bv->bv_ifindex = bvm->ifindex;
			bv->ce_mask |= BRIDGE_VLAN_ATTR_IFINDEX;

			bv->bv_vid = vid;
			bv->ce_mask |= BRIDGE_VLAN_ATTR_VID;

			bv->bv_flags = vinfo->flags;
			bv->ce_mask |= BRIDGE_VLAN_ATTR_FLAGS;

			err = bridge_vlan_set_vlan_attrs(bv, vlan_tb,
							 vid - vinfo->vid);
			if (err) {
				rtnl_bridge_vlan_put(bv);
				return err;
			}

			err = pp->pp_cb((struct nl_object *) bv, pp);
			rtnl_bridge_vlan_put(bv);
			if (err)
				return err;
		}
	}

	return 0;
}

static int bridge_vlan_request_update(struct nl_cache *cache,
				      struct nl_sock *sk)
{
	struct br_vlan_msg *bvm;
	struct nl_msg *msg;
	int err;

	msg = nlmsg_alloc_simple(RTM_GETVLAN, NLM_F_DUMP);
	if (!msg)
		return -NLE_NOMEM;

	err = -NLE_MSGSIZE;
	bvm = nlmsg_reserve(msg, sizeof(struct br_vlan_msg), NLMSG_ALIGNTO);
	if (bvm == NULL)
		goto nla_put_failure;
	bvm->family = AF_BRIDGE;

	NLA_PUT_U32(msg, BRIDGE_VLANDB_DUMP_FLAGS, BRIDGE_VLANDB_DUMPF_STATS);

	err = nl_send_auto(sk, msg);
	if (err > 0)
		err = 0;

nla_put_failure:
	nlmsg_free(msg);
	return err;
}


const char *rtnl_bridge_vlan_port_states[] = {
	[BR_STATE_DISABLED] = "disabled",
	[BR_STATE_LISTENING] = "listening",
	[BR_STATE_LEARNING] = "learning",
	[BR_STATE_FORWARDING] = "forwarding",
	[BR_STATE_BLOCKING] = "blocking",
};

static void bridge_vlan_dump_line(struct nl_object *obj,
				  struct nl_dump_params *dp)
{
	struct rtnl_bridge_vlan *bv = (struct rtnl_bridge_vlan *) obj;

	nl_new_line(dp);

	if (bv->ce_mask & BRIDGE_VLAN_ATTR_IFINDEX) {
		struct nl_cache *link_cache;
		char buf[128];

		link_cache = nl_cache_mngt_require_safe("route/link");
		if (link_cache) {
			nl_dump(dp, "dev %s ",
				rtnl_link_i2name(link_cache, bv->bv_ifindex,
						 buf, sizeof(buf)));
			nl_cache_put(link_cache);
		} else {
			nl_dump(dp, "dev %d ", bv->bv_ifindex);
		}
	}

	if (bv->ce_mask & BRIDGE_VLAN_ATTR_VID)
		nl_dump(dp, "vid %hu ", bv->bv_vid);

	if (bv->ce_mask & BRIDGE_VLAN_ATTR_FLAGS) {
		if (bv->bv_flags & BRIDGE_VLAN_INFO_PVID)
			nl_dump(dp, "ingress-untagged ");
		if (bv->bv_flags & BRIDGE_VLAN_INFO_UNTAGGED)
			nl_dump(dp, "egress-untagged ");
	}

	if (bv->ce_mask & BRIDGE_VLAN_ATTR_STATE) {
		if (bv->bv_state <= BR_STATE_BLOCKING) {
			nl_dump(dp, "state %s ",
				rtnl_bridge_vlan_port_states[bv->bv_state]);
		} else {
			nl_dump(dp, "state (%d) ", bv->bv_state);
		}
	}

	if (bv->ce_mask & BRIDGE_VLAN_ATTR_TUNID)
		nl_dump(dp, "tunid %u ", bv->bv_tunid);

	nl_dump(dp, "\n");
}

static void bridge_vlan_dump_details(struct nl_object *obj,
				     struct nl_dump_params *dp)
{
	bridge_vlan_dump_line(obj, dp);
}

static void bridge_vlan_dump_stats(struct nl_object *obj,
				   struct nl_dump_params *dp)
{
	struct rtnl_bridge_vlan *bv = (struct rtnl_bridge_vlan *) obj;

	bridge_vlan_dump_details(obj, dp);

	nl_dump_line(dp, "    ");

#define BRIDGE_VLAN_SHOW_STAT(name, suffix) \
	if (bv->ce_mask & BRIDGE_VLAN_ATTR_##suffix) \
		nl_dump(dp, #name " %" PRIu64 " ", bv->bv_##name);

	BRIDGE_VLAN_SHOW_STAT(rx_bytes, RX_BYTES);
	BRIDGE_VLAN_SHOW_STAT(rx_packets, RX_PACKETS);
	BRIDGE_VLAN_SHOW_STAT(tx_bytes, TX_BYTES);
	BRIDGE_VLAN_SHOW_STAT(tx_packets, TX_PACKETS);

#undef BRIDGE_VLAN_SHOW_STAT

	nl_dump(dp, "\n");
}

static uint64_t bridge_vlan_compare(struct nl_object *_a,
				    struct nl_object *_b,
				    uint64_t attrs, int flags)
{
	struct rtnl_bridge_vlan *a = (struct rtnl_bridge_vlan *) _a;
	struct rtnl_bridge_vlan *b = (struct rtnl_bridge_vlan *) _b;
	uint64_t diff = 0;

#define BRIDGE_VLAN_DIFF(ATTR, EXPR) \
	ATTR_DIFF(attrs, BRIDGE_VLAN_ATTR_##ATTR, a, b, EXPR)

	diff |= BRIDGE_VLAN_DIFF(IFINDEX, a->bv_ifindex != b->bv_ifindex);
	diff |= BRIDGE_VLAN_DIFF(VID, a->bv_vid != b->bv_vid);
	diff |= BRIDGE_VLAN_DIFF(FLAGS, a->bv_flags != b->bv_flags);
	diff |= BRIDGE_VLAN_DIFF(STATE, a->bv_state != b->bv_state);
	diff |= BRIDGE_VLAN_DIFF(TUNID, a->bv_tunid != b->bv_tunid);

#undef BRIDGE_VLAN_DIFF

	return diff;
}

static void bridge_vlan_keygen(struct nl_object *obj, uint32_t *hashkey,
			       uint32_t table_sz)
{
	struct rtnl_bridge_vlan *bv = (struct rtnl_bridge_vlan *) obj;
	struct {
		uint32_t ifindex;
		uint16_t vid;
	} __attribute__((packed)) bv_key;

	bv_key.ifindex = bv->bv_ifindex;
	bv_key.vid = bv->bv_vid;

	*hashkey = nl_hash(&bv_key, sizeof(bv_key), 0) % table_sz;

	NL_DBG(5, "bridge_vlan %p key (dev %u vid %hu) keysz %zu, hash 0x%x\n",
	       bv, bv_key.ifindex, bv_key.vid, sizeof(bv_key), *hashkey);

	return;
}

static const struct trans_tbl bridge_vlan_attrs[] = {
	__ADD(BRIDGE_VLAN_ATTR_IFINDEX, ifindex),
	__ADD(BRIDGE_VLAN_ATTR_VID, vid),
	__ADD(BRIDGE_VLAN_ATTR_FLAGS, flags),
	__ADD(BRIDGE_VLAN_ATTR_STATE, state),
	__ADD(BRIDGE_VLAN_ATTR_TUNID, tunid),
	__ADD(BRIDGE_VLAN_ATTR_RX_BYTES, rx_bytes),
	__ADD(BRIDGE_VLAN_ATTR_RX_PACKETS, rx_packets),
	__ADD(BRIDGE_VLAN_ATTR_TX_BYTES, tx_bytes),
	__ADD(BRIDGE_VLAN_ATTR_TX_PACKETS, tx_packets),
};

static char *bridge_vlan_attrs2str(int attrs, char *buf, size_t len)
{
	return __flags2str(attrs, buf, len, bridge_vlan_attrs,
			   ARRAY_SIZE(bridge_vlan_attrs));
}

/**
 * @name Get / List
 * @{
 */


/**
 * Allocate bridge_vlan cache and fill in all configured bridge vlans.
 * @arg sk		Netlink socket.
 * @arg result		Pointer to store resulting cache.
 * @arg flags		Flags to set in cache before filling
 *
 * Allocates and initializes a new bridge_vlan cache. If \c sk is valid, a
 * netlink message is sent to the kernel requesting a full dump of all
 * configured bridge vlans. The returned messages are parsed and filled into
 * the cache. If \c sk is NULL, returns 0 but the cache is still empty.
 *
 * @return 0 on success or a negative error code.
 */
int rtnl_bridge_vlan_alloc_cache_flags(struct nl_sock *sk,
				       struct nl_cache **result,
				       unsigned int flags)
{
	struct nl_cache * cache;
	int err;
	
	cache = nl_cache_alloc(&rtnl_bridge_vlan_ops);
	if (!cache)
		return -NLE_NOMEM;

	if (flags)
		nl_cache_set_flags(cache, flags);

	if (sk && (err = nl_cache_refill(sk, cache)) < 0) {
		nl_cache_free(cache);
		return err;
	}

	*result = cache;
	return 0;
}

/**
 * Allocate bridge_vlan cache and fill in all configured bridge vlans.
 * @arg sk		Netlink socket.
 * @arg result		Pointer to store resulting cache.
 *
 * Allocates and initializes a new bridge_vlan cache. If \c sk is valid, a
 * netlink message is sent to the kernel requesting a full dump of all
 * configured bridge vlans. The returned messages are parsed and filled into
 * the cache. If \c sk is NULL, returns 0 but the cache is still empty.
 *
 * @return 0 on success or a negative error code.
 */
int rtnl_bridge_vlan_alloc_cache(struct nl_sock *sk, struct nl_cache **result)
{
	return rtnl_bridge_vlan_alloc_cache_flags(sk, result, 0);
}

/** @} */

/**
 * @name Allocation/Freeing
 * @{
 */

struct rtnl_bridge_vlan *rtnl_bridge_vlan_alloc(void)
{
	return (struct rtnl_bridge_vlan *) nl_object_alloc(
		&bridge_vlan_obj_ops);
}

void rtnl_bridge_vlan_put(struct rtnl_bridge_vlan *bv)
{
	nl_object_put((struct nl_object *) bv);
}

/** @} */

/* only_opts: Only process state and tunnel_info attributes. Only possible
 * 	with RTM_NEWVLAN. Maps to skip_processing in the kernel's
 * 	br_vlan_rtm_process_one(). If unset, also process vlan changes
 * 	(addition and flags).
 */
static int build_bridge_vlan_msg(int cmd, bool only_opts, int tun_cmd,
				 struct br_vlan_msg *bvm,
				 const struct rtnl_bridge_vlan *bv, int flags,
				 struct nl_msg **result)
{
	struct bridge_vlan_info vinfo = {
		.flags = only_opts ? BRIDGE_VLAN_INFO_ONLY_OPTS : 0,
		.vid = bv->bv_vid,
	};
	struct nl_msg *msg;
	struct nlattr *nla, *tun_nla;
	int err;

	BUG_ON(cmd == RTM_DELVLAN && only_opts);

	msg = nlmsg_alloc_simple(cmd, flags);
	if (!msg)
		return -NLE_NOMEM;

	err = nlmsg_append(msg, bvm, sizeof(struct br_vlan_msg), NLMSG_ALIGNTO);
	if (err < 0)
		goto nla_put_failure;

	if (!(nla = nla_nest_start(msg, BRIDGE_VLANDB_ENTRY))) {
		err = -NLE_MSGSIZE;
		goto nla_put_failure;
	}

	if (!only_opts && bv->ce_mask & BRIDGE_VLAN_ATTR_FLAGS)
		vinfo.flags |= bv->bv_flags;

	NLA_PUT(msg, BRIDGE_VLANDB_ENTRY_INFO,
		sizeof(struct bridge_vlan_info), &vinfo);

	if (bv->ce_mask & BRIDGE_VLAN_ATTR_STATE)
		NLA_PUT_U8(msg, BRIDGE_VLANDB_ENTRY_STATE, bv->bv_state);

	if (tun_cmd) {
		tun_nla = nla_nest_start(msg,
					 BRIDGE_VLANDB_ENTRY_TUNNEL_INFO);
		if (!tun_nla) {
			nla_nest_cancel(msg, nla);
			err = -NLE_MSGSIZE;
			goto nla_put_failure;
		}

		if (tun_cmd == RTM_SETLINK) {
			NLA_PUT_U32(msg, BRIDGE_VLANDB_TINFO_ID,
				    bv->bv_tunid);
		}
		NLA_PUT_U32(msg, BRIDGE_VLANDB_TINFO_CMD, tun_cmd);

		nla_nest_end(msg, tun_nla);
	}

	nla_nest_end(msg, nla);

	*result = msg;
	return 0;

nla_put_failure:
	nlmsg_free(msg);
	return err;
}
/**
 * @name Add / Modify
 * @{
 */

/**
 * Build a netlink message requesting the addition of new bridge vlan
 * @arg bv		new bridge_vlan to add
 * @arg flags		additional netlink message flags
 * @arg result		pointer to store resulting netlink message
 *
 * The behaviour of this function is identical to rtnl_bridge_vlan_add() with
 * the exception that it will not send the message but return it in the
 * provided return pointer instead.
 *
 * @see rtnl_bridge_vlan_add()
 *
 * @note This operation is not supported on all kernel versions.
 *
 * @return 0 on success or a negative error code.
 */
int rtnl_bridge_vlan_build_add_request(struct rtnl_bridge_vlan *bv, int flags,
				       struct nl_msg **result)
{
	struct br_vlan_msg bvm = {
		.family = AF_BRIDGE,
		.ifindex = bv->bv_ifindex,
	};
	int tun_cmd = 0;

	if (!(bv->ce_mask & (BRIDGE_VLAN_ATTR_IFINDEX |
			     BRIDGE_VLAN_ATTR_VID))) {
		APPBUG("ifindex and vid must be specified");
		return -NLE_MISSING_ATTR;
	}

	if (bv->ce_mask & BRIDGE_VLAN_ATTR_TUNID)
		tun_cmd = RTM_SETLINK;

	return build_bridge_vlan_msg(RTM_NEWVLAN, false, tun_cmd, &bvm, bv,
				     flags, result);
}

/**
 * Add bridge vlan
 * @arg sk		netlink socket.
 * @arg bv		bridge vlan to add
 * @arg flags		additional netlink message flags
 *
 * Builds a \c RTM_NEWVLAN netlink message requesting the addition of
 * a bridge vlan.
 *
 * After sending, the function will wait for the ACK or an eventual
 * error message to be received and will therefore block until the
 * operation has been completed.
 *
 * @copydoc auto_ack_warning
 *
 * @return 0 on success or a negative error code.
 */
int rtnl_bridge_vlan_add(struct nl_sock *sk, struct rtnl_bridge_vlan *bv,
			 int flags)
{
	struct nl_msg *msg;
	int err;

	if ((err = rtnl_bridge_vlan_build_add_request(bv, flags, &msg)) < 0)
		return err;

	return nl_send_sync(sk, msg);
}

/**
 * Build a netlink message requesting the modification of bridge_vlan
 * @arg orig		original bridge vlan to change
 * @arg changes		bridge vlan containing the changes to be made
 * @arg flags		additional netlink message flags
 * @arg result		pointer to store resulting netlink message
 *
 * The behaviour of this function is identical to rtnl_bridge_vlan_change()
 * with the exception that it will not send the message but return it in the
 * provided return pointer instead.
 *
 * @see rtnl_bridge_vlan_change()
 *
 * @return 0 on success or a negative error code.
 */
int rtnl_bridge_vlan_build_change_request(struct rtnl_bridge_vlan *orig,
					  struct rtnl_bridge_vlan *changes,
					  int flags, struct nl_msg **result)
{
	struct br_vlan_msg bvm = {
		.family = AF_BRIDGE,
		.ifindex = orig->bv_ifindex,
	};
	bool only_opts;
	int tun_cmd;

	if (!(orig->ce_mask & (BRIDGE_VLAN_ATTR_IFINDEX |
			       BRIDGE_VLAN_ATTR_VID)) ||
	    !(changes->ce_mask & (BRIDGE_VLAN_ATTR_IFINDEX |
				  BRIDGE_VLAN_ATTR_VID))) {
		APPBUG("ifindex and vid must be specified");
		return -NLE_MISSING_ATTR;
	}

	if (changes->bv_ifindex != orig->bv_ifindex) {
		APPBUG("bridge_vlan change: ifindex is immutable");
		return -NLE_IMMUTABLE;
	}

	if (changes->bv_vid != orig->bv_vid) {
		APPBUG("bridge_vlan change: vid is immutable");
		return -NLE_IMMUTABLE;
	}

	if (changes->ce_mask & BRIDGE_VLAN_ATTR_FLAGS)
		only_opts = false;
	else
		only_opts = true;

	if (!(orig->ce_mask & BRIDGE_VLAN_ATTR_TUNID) &&
	    changes->ce_mask & BRIDGE_VLAN_ATTR_TUNID) {
		tun_cmd = RTM_SETLINK;
	} else if (orig->ce_mask & BRIDGE_VLAN_ATTR_TUNID &&
		   !(changes->ce_mask & BRIDGE_VLAN_ATTR_TUNID)) {
		tun_cmd = RTM_DELLINK;
	} else {
		tun_cmd = 0;
	}

	return build_bridge_vlan_msg(RTM_NEWVLAN, only_opts, tun_cmd, &bvm,
				     changes, flags, result);
}

/**
 * Change bridge_vlan
 * @arg sk		netlink socket.
 * @arg orig		original bridge_vlan to be changed
 * @arg changes		bridge_vlan containing the changes to be made
 * @arg flags		additional netlink message flags
 *
 * Builds a \c RTM_NEWVLAN netlink message requesting the change of
 * a bridge vlan.
 *
 * After sending, the function will wait for the ACK or an eventual
 * error message to be received and will therefore block until the
 * operation has been completed.
 *
 * @copydoc auto_ack_warning
 *
 * @return 0 on success or a negative error code.
 */
int rtnl_bridge_vlan_change(struct nl_sock *sk, struct rtnl_bridge_vlan *orig,
			    struct rtnl_bridge_vlan *changes, int flags)
{
	struct nl_msg *msg;
	int err;
	
	if ((err = rtnl_bridge_vlan_build_change_request(orig, changes, flags,
							 &msg)) < 0)
		return err;

	return nl_send_sync(sk, msg);
}

/** @} */

/**
 * @name Delete
 * @{
 */

/**
 * Build a netlink message requesting the deletion of a bridge_vlan
 * @arg bridge_vlan     bridge_vlan to delete
 * @arg result		Pointer to store resulting netlink message
 *
 * The behaviour of this function is identical to rtnl_bridge_vlan_delete()
 * with the exception that it will not send the message but return it in the
 * provided return pointer instead.
 *
 * @see rtnl_bridge_vlan_delete()
 *
 * @return 0 on success or a negative error code.
 */
int rtnl_bridge_vlan_build_delete_request(const struct rtnl_bridge_vlan *bv,
					  struct nl_msg **result)
{
	struct br_vlan_msg bvm = {
		.family = AF_BRIDGE,
		.ifindex = bv->bv_ifindex,
	};

	if (!(bv->ce_mask & (BRIDGE_VLAN_ATTR_IFINDEX |
			     BRIDGE_VLAN_ATTR_VID))) {
		APPBUG("ifindex and vid must be specified");
		return -NLE_MISSING_ATTR;
	}

	return build_bridge_vlan_msg(RTM_DELVLAN, false, 0, &bvm, bv, 0,
				     result);
}

/**
 * Delete bridge_vlan
 * @arg sk		Netlink socket
 * @arg bridge_vlan     bridge_vlan to delete
 *
 * Builds a \c RTM_DELVLAN netlink message requesting the deletion of
 * a bridge vlan which has been previously added to the kernel and
 * sends the message to the kernel.
 *
 * If no matching bridge_vlan exists, the function will return
 * -NLE_OBJ_NOTFOUND.
 *
 * After sending, the function will wait for the ACK or an eventual
 * error message to be received and will therefore block until the
 * operation has been completed.
 *
 * @copydoc auto_ack_warning
 *
 * @return 0 on success or a negative error code.
 */
int rtnl_bridge_vlan_delete(struct nl_sock *sk,
			    const struct rtnl_bridge_vlan *bv)
{
	struct nl_msg *msg;
	int err;
	
	if ((err = rtnl_bridge_vlan_build_delete_request(bv, &msg)) < 0)
		return err;

	return nl_send_sync(sk, msg);
}

/** @} */

/**
 * @name Attribute Modification
 * @{
 */

void rtnl_bridge_vlan_set_ifindex(struct rtnl_bridge_vlan *bv, int ifindex)
{
	bv->bv_ifindex = ifindex;
	bv->ce_mask |= BRIDGE_VLAN_ATTR_IFINDEX;
}

int rtnl_bridge_vlan_get_ifindex(struct rtnl_bridge_vlan *bv)
{
	return bv->bv_ifindex;
}

void rtnl_bridge_vlan_set_vid(struct rtnl_bridge_vlan *bv, uint16_t vid)
{
	bv->bv_vid = vid;
	bv->ce_mask |= BRIDGE_VLAN_ATTR_VID;
}

uint16_t rtnl_bridge_vlan_get_vid(struct rtnl_bridge_vlan *bv)
{
	return bv->bv_vid;
}

void rtnl_bridge_vlan_set_flags(struct rtnl_bridge_vlan *bv, uint16_t flags)
{
	bv->bv_flags = flags;
	bv->ce_mask |= BRIDGE_VLAN_ATTR_FLAGS;
}

uint16_t rtnl_bridge_vlan_get_flags(struct rtnl_bridge_vlan *bv)
{
	return bv->bv_flags;
}

void rtnl_bridge_vlan_set_state(struct rtnl_bridge_vlan *bv, uint8_t state)
{
	bv->bv_state = state;
	bv->ce_mask |= BRIDGE_VLAN_ATTR_STATE;
}

uint8_t rtnl_bridge_vlan_get_state(struct rtnl_bridge_vlan *bv)
{
	return bv->bv_state;
}

/**
 * Get Spanning Tree Protocol (STP) port state
 * @arg link		Link object
 *
 * This specific API was requested in order to have a function that can be
 * easily integrated into switchd code that previously used
 * rtnl_link_bridge_get_port_state().
 *
 * @return The STP port state or a negative error code.
 */
int rtnl_link_bridge_vlan_get_port_state(struct rtnl_link *link, uint16_t vid)
{
	struct rtnl_bridge_vlan *needle, *match;
	struct nl_cache *cache;
	int result;

	cache = nl_cache_mngt_require_safe("route/bridge_vlan");
	if (!cache) {
		return -NLE_NOCACHE;
	}

	needle = rtnl_bridge_vlan_alloc();
	rtnl_bridge_vlan_set_ifindex(needle, rtnl_link_get_ifindex(link));
	rtnl_bridge_vlan_set_vid(needle, vid);

	match = (struct rtnl_bridge_vlan *)nl_cache_find(cache,
							 OBJ_CAST(needle));
	rtnl_bridge_vlan_put(needle);
	if (match) {
		result = rtnl_bridge_vlan_get_state(match);
	} else {
		result = -NLE_OBJ_NOTFOUND;
	}

	rtnl_bridge_vlan_put(match);
	nl_cache_put(cache);

	return result;
}

void rtnl_bridge_vlan_set_tunid(struct rtnl_bridge_vlan *bv, uint32_t tunid)
{
	bv->bv_tunid = tunid;
	bv->ce_mask |= BRIDGE_VLAN_ATTR_TUNID;
}

uint32_t rtnl_bridge_vlan_get_tunid(struct rtnl_bridge_vlan *bv)
{
	return bv->bv_tunid;
}

/** @} */

static struct nl_object_ops bridge_vlan_obj_ops = {
	.oo_name		= "route/bridge_vlan",
	.oo_size		= sizeof(struct rtnl_bridge_vlan),
	.oo_dump = {
	    [NL_DUMP_LINE]	= bridge_vlan_dump_line,
	    [NL_DUMP_DETAILS]	= bridge_vlan_dump_details,
	    [NL_DUMP_STATS]	= bridge_vlan_dump_stats,
	},
	.oo_compare		= bridge_vlan_compare,
	.oo_keygen		= bridge_vlan_keygen,
	.oo_attrs2str		= bridge_vlan_attrs2str,
	.oo_id_attrs		= BRIDGE_VLAN_ATTR_IFINDEX |
		BRIDGE_VLAN_ATTR_VID,
};

static struct nl_af_group bridge_vlan_groups[] = {
	{ AF_BRIDGE, RTNLGRP_BRVLAN },
	{ END_OF_GROUP_LIST },
};

static struct nl_cache_ops rtnl_bridge_vlan_ops = {
	.co_name		= "route/bridge_vlan",
	.co_hdrsize		= sizeof(struct br_vlan_msg),
	.co_msgtypes		= {
					{ RTM_NEWVLAN, NL_ACT_NEW, "new" },
					{ RTM_DELVLAN, NL_ACT_DEL, "del" },
					{ RTM_GETVLAN, NL_ACT_GET, "get" },
					END_OF_MSGTYPES_LIST,
				  },
	.co_protocol		= NETLINK_ROUTE,
	.co_groups		= bridge_vlan_groups,
	.co_request_update	= bridge_vlan_request_update,
	.co_msg_parser		= bridge_vlan_msg_parser,
	.co_obj_ops		= &bridge_vlan_obj_ops,
	.co_hash_size           = 16384,
};

static void __init bridge_vlan_init(void)
{
	nl_cache_mngt_register(&rtnl_bridge_vlan_ops);
}

static void __exit bridge_vlan_exit(void)
{
	nl_cache_mngt_unregister(&rtnl_bridge_vlan_ops);
}

/** @} */
