/*
 *	This library is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU Lesser General Public
 *	License as published by the Free Software Foundation version 2.1
 *	of the License.
 *
 * Copyright (c) 2021 Cumulus Networks <dev-support@cumulusnetworks.com>
 */

#include <netlink/route/lwt/api.h>
#include <endian.h>
#include <errno.h>
#include <inttypes.h>
#include <linux/lwtunnel.h>
#include <netlink-private/netlink.h>
#include <netlink/attr.h>
#include <netlink/route/lwt/ip6.h>

static struct nla_policy lwt_ip6_policy[LWTUNNEL_IP6_MAX + 1] = {
	[LWTUNNEL_IP6_ID]		= { .type = NLA_U64 },
	[LWTUNNEL_IP6_DST]		= { .minlen = sizeof(struct in6_addr) },
	[LWTUNNEL_IP6_SRC]		= { .minlen = sizeof(struct in6_addr) },
	[LWTUNNEL_IP6_HOPLIMIT]		= { .type = NLA_U8 },
	[LWTUNNEL_IP6_TC]		= { .type = NLA_U8 },
	[LWTUNNEL_IP6_FLAGS]		= { .type = NLA_U16 },
};

struct rtnl_lwt_ip6_encap *rtnl_lwt_ip6_encap_hdr(struct rtnl_lwt_encap *encap)
{
	return (struct rtnl_lwt_ip6_encap *)encap->l_hdr;
}

static int lwt_ip6_build_encap(struct nlattr *nla,
			       struct rtnl_lwt_encap **encap)
{
	struct nlattr *tb[LWTUNNEL_IP6_MAX + 1];
	struct rtnl_lwt_encap *newencap;
	struct rtnl_lwt_ip6_encap *ip6_encap;
	int err;

	err = nla_parse_nested(tb, LWTUNNEL_IP6_MAX, nla, lwt_ip6_policy);
	if (err < 0)
		goto out_err;

	newencap = malloc(sizeof(struct rtnl_lwt_encap) +
			  sizeof(struct rtnl_lwt_ip6_encap));
	if (!newencap) {
		err = -ENOMEM;
		goto out_err;
	}

	newencap->l_type = LWTUNNEL_ENCAP_IP6;
	newencap->l_len = sizeof(struct rtnl_lwt_ip6_encap);

	ip6_encap = rtnl_lwt_ip6_encap_hdr(newencap);
	ip6_encap->li6_id = be64toh(nla_get_u64(tb[LWTUNNEL_IP6_ID]));
	ip6_encap->li6_dst = nl_addr_alloc_attr(tb[LWTUNNEL_IP6_DST], AF_INET6);
	if (!ip6_encap->li6_dst) {
		free(newencap);
		err = -ENOMEM;
		goto out_err;
	}

	*encap = newencap;
	return 0;

out_err:
	*encap = NULL;
	return err;
}

static struct rtnl_lwt_encap *lwt_ip6_clone_encap(struct rtnl_lwt_encap *src)
{
	struct rtnl_lwt_ip6_encap *src_encap, *dst_encap;
	struct rtnl_lwt_encap *newencap;

	newencap = malloc(sizeof(struct rtnl_lwt_encap) +
			  sizeof(struct rtnl_lwt_ip6_encap));
	if (!newencap)
		return NULL;

	newencap->l_type = LWTUNNEL_ENCAP_IP6;
	newencap->l_len = sizeof(struct rtnl_lwt_ip6_encap);

	src_encap = rtnl_lwt_ip6_encap_hdr(src);
	dst_encap = rtnl_lwt_ip6_encap_hdr(newencap);
	dst_encap->li6_id = src_encap->li6_id;
	dst_encap->li6_dst = nl_addr_clone(src_encap->li6_dst);
	if (!dst_encap->li6_dst) {
		free(newencap);
		return NULL;
	}

	return newencap;
}

static void lwt_ip6_release_encap(struct rtnl_lwt_encap *encap)
{
	struct rtnl_lwt_ip6_encap *ip6_encap;

	if (!encap)
		return;

	ip6_encap = rtnl_lwt_ip6_encap_hdr(encap);
	nl_addr_put(ip6_encap->li6_dst);
	free(encap);
}

static int lwt_ip6_compare_encap(struct rtnl_lwt_encap *a,
				 struct rtnl_lwt_encap *b)
{
	struct rtnl_lwt_ip6_encap *ip6_encap_a = rtnl_lwt_ip6_encap_hdr(a);
	struct rtnl_lwt_ip6_encap *ip6_encap_b = rtnl_lwt_ip6_encap_hdr(b);

	if (a->l_type != b->l_type || a->l_len != b->l_len)
		return 1;

	if (ip6_encap_a->li6_id != ip6_encap_b->li6_id)
		return 1;

	return nl_addr_cmp(ip6_encap_a->li6_dst, ip6_encap_b->li6_dst);
}

static void lwt_ip6_dump_encap(struct rtnl_lwt_encap *encap,
			       struct nl_dump_params *p)
{
	const struct rtnl_lwt_ip6_encap *ip6_encap;
	char buf[64];

       	ip6_encap = rtnl_lwt_ip6_encap_hdr(encap);
	nl_dump(p, "encap ip6 id %" PRIu64 " dst %s ", ip6_encap->li6_id,
		nl_addr2str(ip6_encap->li6_dst, buf, sizeof(buf)));
}

int rtnl_lwt_is_ip6(struct rtnl_lwt_encap *encap)
{
	return encap->l_type == LWTUNNEL_ENCAP_IP6;
}

static struct rtnl_lwt_ops lwt_ip6_ops = {
	.lwt_type		= LWTUNNEL_ENCAP_IP6,
	.lwt_build_encap	= &lwt_ip6_build_encap,
	.lwt_clone_encap	= &lwt_ip6_clone_encap,
	.lwt_release_encap	= &lwt_ip6_release_encap,
	.lwt_compare_encap	= &lwt_ip6_compare_encap,
	.lwt_dump_encap		= &lwt_ip6_dump_encap,
};

static void __init lwt_ip6_init(void)
{
	rtnl_lwt_ops_register(&lwt_ip6_ops);
}

static void __exit lwt_ip6_exit(void)
{
	rtnl_lwt_ops_unregister(&lwt_ip6_ops);
}
