#include <linux/if_bridge.h>
#include <netlink/cli/utils.h>
#include <netlink/cli/link.h>
#include <netlink/route/bridge_vlan.h>

static struct nl_sock *sock;
static int quiet = 0;

static struct vlan_flag {
	uint16_t value;
	const char *name;
} vlan_flags[] = {
	{ BRIDGE_VLAN_INFO_PVID, "ingress-untagged" },
	{ BRIDGE_VLAN_INFO_UNTAGGED, "egress-untagged" },
	{ 0, NULL },
};

static void usage(FILE *stream)
{
	int i;

	fprintf(stream,
		"Usage: nl-bridge-vlan [OPTION]... <DEV OPTION...> <VID>\n"
		"\n"
		"Options\n"
		" -q, --quiet           Do not print informal notifications.\n"
		" -h, --help            Show this help.\n"
		" -v, --version         Show versioning information.\n"
		"\n"
		"Dev Options\n"
		" -n, --name=NAME       interface name\n"
		" -i, --index=INDEX     interface index\n"
		"Change Options\n"
		" -f, --flag=FLAG       set vlan flag.\n"
		" -F, --no-flag=FLAG    unset vlan flag.\n"
		"                       Possible flags are:\n"
	       );
	for (i = 0; vlan_flags[i].name; i++) {
		fprintf(stream, "                          %s\n",
			vlan_flags[i].name);
	}
	fprintf(stream,
		" -s, --state=STATE     set vlan state. Possible states are:\n"
	       );
	for (i = 0; i <= BR_STATE_BLOCKING; i++) {
		fprintf(stream, "                          %s\n",
			rtnl_bridge_vlan_port_states[i]);
	}
	fprintf(stream,
		" -t, --tunnel=TUNID    set vlan tunnel id.\n"
		" -T, --no-tunnel       unset vlan tunnel id.\n"
		" -d, --delete          delete vlan.\n"
	       );
}

static int parse_flag(const char *name, uint16_t *mask)
{
	struct vlan_flag *flag, *match = NULL;

	for (flag = &vlan_flags[0]; flag->name; flag++) {
		size_t len;

		len = strlen(name);
		if (strncmp(name, flag->name, len) == 0) {
			if (!match) {
				match = flag;
			} else {
				return 2;
			}
		}
	}

	if (match) {
		*mask |= match->value;
		return 1;
	}
	return 0;
}

#define TUN_ACT_NONE 0
#define TUN_ACT_SET 1
#define TUN_ACT_UNSET 2

int main(int argc, char *argv[])
{
	uint16_t flags, flags_set = 0, flags_unset = 0, *flags_group;
	struct nl_cache *link_cache, *bv_cache;
	struct rtnl_bridge_vlan *cur, *desired;
	bool cur_required, delete_vlan = false;
	struct rtnl_link *needle, *match;
	struct nl_dump_params params = {
		.dp_type = NL_DUMP_LINE,
		.dp_prefix = 4,
		.dp_fd = stdout,
	};
	unsigned long vid, tunid = 0;
	int tun_act = TUN_ACT_NONE;
	int retval, dev_opt_nb = 0;
	char *endptr;

	sock = nl_cli_alloc_socket();
	nl_cli_connect(sock, NETLINK_ROUTE);
	link_cache = nl_cli_link_alloc_cache(sock);

	retval = rtnl_bridge_vlan_alloc_cache(sock, &bv_cache);
	if (retval < 0) {
		nl_cli_fatal(retval, "Unable to allocate bridge_vlan cache: %s",
			     nl_geterror(retval));
	}
	nl_cache_mngt_provide(bv_cache);

	needle = nl_cli_link_alloc();
	desired = rtnl_bridge_vlan_alloc();

	while (true) {
		static struct option longopts[] = {
			{ "quiet", 0, 0, 'q' },
			{ "help", 0, 0, 'h' },
			{ "version", 0, 0, 'v' },
			{ "name", 1, 0, 'n' },
			{ "index", 1, 0, 'i' },
			{ "flag", 1, 0, 'f' },
			{ "no-flag", 1, 0, 'F' },
			{ "state", 1, 0, 's' },
			{ "tunnel", 1, 0, 't' },
			{ "no-tunnel", 0, 0, 'T' },
			{ "delete", 0, 0, 'd' },
			{ 0, 0, 0, 0 }
		};
		bool found = false;
		int i, c;

		c = getopt_long(argc, argv, "qhvn:i:f:F:s:t:Td", longopts,
				NULL);
		if (c == -1)
			break;

		flags_group = NULL;

		switch (c) {
		case 'q':
			quiet = 1;
			break;
		case 'h':
			usage(stdout);
			exit(0);
			break;
		case 'v':
			nl_cli_print_version();
			break;
		case 'n':
			dev_opt_nb++;
			nl_cli_link_parse_name(needle, optarg);
			break;
		case 'i':
			dev_opt_nb++;
			nl_cli_link_parse_ifindex(needle, optarg);
			break;
		case 'f':
			flags_group = &flags_set;
			/* fall-through */
		case 'F':
			if (!flags_group) {
				flags_group = &flags_unset;
			}

			if (parse_flag(optarg, flags_group) != 1) {
				fprintf(stderr, "Error: unknown or ambiguous vlan flag \"%s\"\n",
					optarg);
				usage(stderr);
				exit(1);
			}
			break;
		case 's':
			for (i = 0; i <= BR_STATE_BLOCKING; i++) {
				if (!strcmp(optarg,
					    rtnl_bridge_vlan_port_states[i])) {
					rtnl_bridge_vlan_set_state(desired, i);
					found = true;
					break;
				}
			}
			if (!found) {
				fprintf(stderr, "Error: unknown vlan state \"%s\"\n",
					optarg);
				usage(stderr);
				exit(1);
			}
			delete_vlan = false;
			break;
		case 't':
			tun_act = TUN_ACT_SET;
			tunid = strtoul(optarg, &endptr, 0);
			if (*endptr != '\0') {
				fprintf(stderr, "Error: could not interpret tunid \"%s\" as an integer\n",
					optarg);
				usage(stderr);
				exit(1);
			}
			delete_vlan = false;
			break;
		case 'T':
			tun_act = TUN_ACT_UNSET;
			delete_vlan = false;
			break;
		case 'd':
			tun_act = TUN_ACT_NONE;
			delete_vlan = true;
			break;
		}
	}

	if (!dev_opt_nb) {
		fprintf(stderr, "Error: missing dev option\n");
		usage(stderr);
		exit(1);
	}

	if (optind >= argc) {
		fprintf(stderr, "Error: missing vid argument\n");
		usage(stderr);
		exit(1);
	}

	if (flags_set & flags_unset) {
		fprintf(stderr, "Error: conflicting vlan flag actions\n");
		usage(stderr);
		exit(1);
	}

	vid = strtoul(argv[optind], &endptr, 0);
	if (*endptr != '\0') {
		fprintf(stderr, "Error: could not interpret vid \"%s\" as an integer\n",
			argv[optind]);
		usage(stderr);
		exit(1);
	}
	rtnl_bridge_vlan_set_vid(desired, vid);
	optind++;

	if (optind < argc) {
		fprintf(stderr, "Error: too many arguments\n");
		usage(stderr);
		exit(1);
	}

	match = (struct rtnl_link *)nl_cache_find(link_cache,
						  OBJ_CAST(needle));
	nl_cache_mngt_unprovide(link_cache);
	nl_cache_free(link_cache);
	if (match) {
		rtnl_bridge_vlan_set_ifindex(desired,
					     rtnl_link_get_ifindex(match));
		rtnl_link_put(match);
	} else {
		fprintf(stderr, "Error: interface not found\n");
		exit(1);
	}
	rtnl_link_put(needle);

	if (flags_unset || tun_act == TUN_ACT_UNSET || delete_vlan) {
		cur_required = true;
	} else {
		cur_required = false;
	}

	cur = (struct rtnl_bridge_vlan *)nl_cache_search(bv_cache,
							 OBJ_CAST(desired));
	if (cur_required && !cur) {
		fprintf(stderr, "Error: current bridge_vlan not found\n");
		exit(1);
	}
	nl_cache_mngt_unprovide(bv_cache);
	nl_cache_free(bv_cache);
	if (cur && !quiet) {
		printf("Current bridge_vlan:\n");
		nl_object_dump(OBJ_CAST(cur), &params);
	}

	if (cur) {
		flags = rtnl_bridge_vlan_get_flags(cur);
	} else {
		flags = 0;
	}
	flags = (flags & ~flags_unset) | flags_set;
	rtnl_bridge_vlan_set_flags(desired, flags);

	if (delete_vlan) {
		retval = rtnl_bridge_vlan_delete(sock, cur);
	} else if (tun_act == TUN_ACT_UNSET) {
		retval = rtnl_bridge_vlan_change(sock, cur, desired, 0);
	} else {
		if (tun_act == TUN_ACT_SET) {
			rtnl_bridge_vlan_set_tunid(desired, tunid);
		}
		retval = rtnl_bridge_vlan_add(sock, desired, 0);
	}
	if (retval < 0) {
		nl_cli_fatal(retval, "Unable to change bridge_vlan: %s",
			     nl_geterror(retval));
	}
	rtnl_bridge_vlan_put(cur);

	if (!quiet) {
		printf("New bridge_vlan:\n");
		nl_object_dump(OBJ_CAST(desired), &params);
	}

	rtnl_bridge_vlan_put(desired);
	nl_socket_free(sock);

	return 0;
}
